# namadu-0.0 - distributed namazu
#
# $Id: namadu.rb,v 1.6 2000/05/31 16:03:05 mas Exp $
# Copyright (c) 2000 Masatoshi SEKI
#
# namadu.rb is copyrighted free software by Masatoshi SEKI.
# You can redistribute it and/or modify it under the same term as Ruby.

require 'rbnamazu'

class Namadu
  class NamazuFieldLoader < NamazuFieldAll
    def gets_as_array(docid, fieldnames = nil)
      unless fieldnames
	fieldnames = @fieldnames
      end
      fieldnames.collect do |fieldname|
	field = @nmzfieldall[fieldname].gets(docid)
	field.chomp!()
	field
      end
    end

    def get_field(docid, fieldname)
      return nil unless fieldnames.include?(fieldname)
      v = @nmzfieldall[fieldname].gets(docid)
      v.chomp
    end
  end

  class Doc
    def initialize(index, docid, score, time)
      @index = index
      @docid = docid
      @score = score
      @time = time
      @fields = nil
      @sortkey = nil
    end
    attr_reader(:index, :docid, :score, :time, :sortkey)

    public
    def fields
      return @fields if @fields

      @fields = @index.fields_loader.gets(@docid, nil)
      @fields['score'] = @score
      @fields['time'] = @time
      @fields
    end

    public
    def load_sortkey(field)
      @sortkey = self[field]
    end

    public
    def [](field)
      case field
      when 'score'
	@score
      when 'time'
	@time
      else
	@index.fields_loader.get_field(@docid, field)
      end
    end
  end

  class Result
    def initialize(list, words)
      @ref_hits = Hash.new(0)
      @docs = []
      list.each do |idx|
	r = idx.find(words)
	r.ref_hits.each do |k,v|
	  @ref_hits[k] += v
	end
	@docs.concat(r.docs)
      end
      @num_docs = @docs.size
    end
    attr_reader(:ref_hits, :num_docs, :docs)

    public
    def sort!(order='score', ascending=false)
      case order
      when 'score'
	@docs.sort! { |a, b| b.score <=> a.score }
      when 'early'
	@docs.sort! { |a, b| a.time <=> a.time }	
      when 'late', 'date'
	@docs.sort! { |a, b| b.time <=> a.time }	
      when /^field:(.+)$/
	key = $1		# FIXME!
	@docs.each do |doc| 
	  doc.load_sortkey(key) 
	end
	@docs.sort! { |a, b| b.sortkey <=> a.sortkey }	
      end
      @docs.reverse! if ascending
    end

    private
    def summary(doc, fields)
      if fields
	fields.collect do |f|
	  doc[f]
	end
      else
	doc.fields
      end
    end

    public
    def result(fields, max=20, whence=0)
      docs = (max) ? @docs[whence, max] : @docs
      if iterator?
	docs.each do |doc|
	  yield(summary(doc, fields))
	end
      else
	docs.collect do |doc|
	  summary(doc, fields)
	end
      end
    end
  end

  class Index
    class Result1
      def initialize(index, words)
	scores, times, ref = index.finder.get_scores(words)

	@ref_hits = make_ref_hits(words, ref)
	@docs = []
	scores.each do |k, v|
	  @docs.push(Doc.new(index, k, v, times[k]))
	end
	@num_docs = @docs.size
      end
      attr_reader(:ref_hits, :num_docs, :docs)
      
      private
      def make_ref_hits(words, counts)
	hash = {}
	words.each_with_index do |w, i|
	  hash[w] = counts[i]	
	end
	hash.freeze
      end
    end

    def initialize(index)
      @index = index.dup.freeze
      @finder = NamazuScores.new(@index)
      @fields_loader = NamazuFieldLoader.new(@index)
    end
    attr_reader(:finder, :fields_loader)

    def find(words)
      Result1.new(self, words)
    end
  end

  def initialize(index_list)
    @list = index_list.collect {|idx| Index.new(idx)}
  end

  public
  def search(words, order='score')
    words = words.collect { |w| w.downcase }      
    r = Result.new(@list, words)
    r.sort!(order) if order
    yield(r) if iterator? 		# GC trick
    r
  end

  public
  def query(words, order='score', fields=nil, max=20, whence=0)
    r = search(words, order)
    list = r.result(fields, max, whence)
    [list, r.num_docs, r.ref_hits]
  end
end

require 'drb/drb'

class Namadu
  include DRbUndumped
  class Result
    include DRbUndumped
  end
end

if __FILE__ == $0
  if ARGV[0] == '--drb'
    ARGV.shift
    index = ARGV.shift.split
    nmd = Namadu.new(index)
    DRb.start_service(ARGV.shift || 'druby://:7952', nmd)
    puts DRb.uri
    if $DEBUG
      gets
    else
      DRb.thread.join
    end
  else
    words = ARGV.shift.split
    nmd = Namadu.new(ARGV)
    r = nmd.search(words, 'field:subject')
    r.result(%w(subject uri score)) do |rec| p rec end
  end
end
