
# numa - distributed namazu client
#
# $Id: numa.rb,v 1.2 2000/06/01 15:52:46 mas Exp $
# Copyright (c) 2000 Masatoshi SEKI
#
# numa.rb is copyrighted free software by Masatoshi SEKI.
# You can redistribute it and/or modify it under the same term as Ruby.

require 'thread'

class Numa
  def initialize(nmd, words, block=20)
    @nmz = nmd
    @block = block
    @pipe = SizedQueue.new(block*2)
    @thread = main_thread(words)
    @num_docs = @pipe.pop
    @eof = (@length == 0)
    @keep = nil
  end
  public
  attr_reader(:num_docs)

  private
  def main_thread(words)
    Thread.new do
      main_loop(words)
    end
  end

  private
  def main_loop(words)
    list = nil
    whence = 0
    @nmz.search(words, "late") do |r|
      sz = r.num_docs
      @pipe.push sz
      while sz > 0
	list = r.result(nil, @block, whence)
	whence += list.size
	sz -= list.size
	list.each do |e|
	  @pipe.push(e)
	end
      end
      @pipe.push(nil)
    end
  end

  public
  def get
    if @keep
      r = @keep
      @keep = nil
    else
      r = @pipe.pop
    end
    @eof = true unless r
    r
  end

  private
  def unget(r)
    raise 'too many unget' if @keep
    @keep = r
  end

  public
  def peek
    r = get
    unget(r)
    r
  end

  public
  def close
    @thread.kill
    @thread = nil
    @pipe = nil
  end
  
  public
  def eof?
    @eof
  end
end

class Numas
  def initialize(words, list, block=20)
    @pool = []
    numa_create(list, words, block)
    @num_docs = 0
    @pool.each do |numa|
      @num_docs += numa.num_docs
    end
    @eof = false
  end
  public
  attr_reader(:num_docs)

  private
  def numa_create(list, words, block)
    queue = Queue.new
    list.each do |nmz|
      tmp = nmz
      Thread.new do
	queue.push(Numa.new(tmp, words, block))
      end
    end
    list.each do |nmz|
      @pool.push(queue.pop)
    end
  end
  
  public
  def get
    return nil if @eof
    @pool.delete_if do |numa|
      numa.eof?
    end
    time = max = nil
    maxidx = nil
    @pool.each_with_index do |numa, idx|
      r = numa.peek
      next unless r
      time = r['time']
      max = time unless max
      maxidx = idx unless maxidx
      if max < time
	max = time
	maxidx = idx
      end
    end
    unless maxidx
      @eof = true
      return nil
    end
    @pool[maxidx].get
  end
end

if __FILE__ == $0
  def usage
    $stderr.puts("usage: #{$0} <'keyword...'> <server_uri1> <server_uri2> <..>")
    exit 1
  end

  usage if ARGV.size <= 1

  words = ARGV.shift.split

  require 'drb/drb'
  DRb.start_service()
  ro_list = ARGV.collect { |uri| DRbObject.new(nil, uri) }

  numas = Numas.new(words, ro_list, 20)
  p numas.num_docs

  while r = numas.get
#    sleep 1
    puts r['uri']
    puts r['date']
    puts r['summary']
    puts
  end
end


